@@ -1,3 +1,8 @@
+# Copyright 2021 DeepL GmbH (https://www.deepl.com)
+# Use of this source code is governed by an MIT
+# license that can be found in the LICENSE file.
+
+
 class DeepLException(Exception):
     """Base class for deepl module exceptions."""
 
@@ -43,3 +48,9 @@ class DocumentTranslationException(DeepLException):
 
     def __str__(self):
         return f"{super()}, document request: {self.document_request}"
+
+
+class GlossaryNotFoundException(DeepLException):
+    """The specified glossary was not found."""
+
+    pass
